"""Main."""
from __future__ import unicode_literals
import sys
import argparse
from pyspelling import spellcheck, version


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='spellcheck', description='Spell checking tool.')
    # Flag arguments
    parser.add_argument('--version', action='version', version=('%(prog)s ' + version))
    parser.add_argument('--verbose', '-v', action='count', default=0, help="Verbosity level.")
    parser.add_argument('--name', '-n', action='store', default='', help="Specific spelling task by name to run.")
    parser.add_argument('--binary', '-b', action='store', default='', help="Provide path to spell checker's binary.")
    parser.add_argument('--config', '-c', action='store', default='.spelling.yml', help="Spelling config.")
    parser.add_argument(
        '--spellchecker', '-s', action='store', default='aspell', help="Choose between aspell and hunspell"
    )
    args = parser.parse_args()

    return run(args.config, args.name, args.binary, args.verbose, args.spellchecker)


def run(config, name='', binary='', verbose=0, spellchecker='aspell'):
    """Run."""

    fail = False
    count = 0
    for results in spellcheck(config, name, binary, verbose, spellchecker):
        count += 1
        if results.error:
            fail = True
            print('ERROR: Could not read %s -- %s' % (results.context, results.error))
        elif results.words:
            fail = True
            print('Misspelled words:\n<%s> %s' % (results.category, results.context))
            print('-' * 80)
            for word in results.words:
                print(word)
            print('-' * 80)
            print('')
        elif verbose:
            print('CHECK: %s' % results.context)

    return fail


if __name__ == "__main__":
    sys.exit(main())



"""JavaScript filter."""
import re
import codecs
from .. import filters
import textwrap

RE_LINE_PRESERVE = re.compile(r"\r?\n", re.MULTILINE)
RE_JSDOC = re.compile(r"(?s)^/\*\*$(.*?)[ \t]*\*/", re.MULTILINE)
RE_COMMENT = re.compile(
    r'''(?x)
        (?P<comments>
            (?P<block>/\*[^*]*\*+(?:[^/*][^*]*\*+)*/)                        # multi-line comments
          | (?P<start>^)?(?P<leading_space>[ \t]*)?(?P<line>//(?:[^\r\n])*)  # single line comments
        )
      | (?P<code>
            "(?:\\.|[^"\\])*"                                                # double quotes
            '(?:\\.|[^'\\])*'                                                # single quotes
          | .[^/"']*?                                                        # everything else
        )
    ''',
    re.DOTALL | re.MULTILINE
)


class JavaScriptFilter(filters.Filter):
    """JavaScript filter."""

    def __init__(self, options, default_encoding='ascii'):
        """Initialization."""

        self.blocks = options.get('block_comments', True) is True
        self.lines = options.get('line_comments', True) is True
        # self.strings = options.get('strings', False) is True
        self.group_comments = options.get('group_comments', False) is True
        self.jsdocs = options.get('jsdocs', False) is True
        super(JavaScriptFilter, self).__init__(options, default_encoding)

    def _evaluate(self, m):
        """Search for comments."""

        g = m.groupdict()
        if g["code"]:
            text = g["code"]
            self.line_num += text.count('\n')
        else:
            if g['block']:
                if self.jsdocs:
                    m1 = RE_JSDOC.match(g['comments'])
                    if m1:
                        lines = []
                        for line in m1.group(1).splitlines(True):
                            l = line.lstrip()
                            lines.append(l[1:] if l.startswith('*') else l)
                        self.jsdoc_comments.append([''.join(lines), self.line_num])
                    elif self.blocks:
                        self.block_comments.append([g['block'][2:-2], self.line_num])
                elif self.blocks:
                    self.block_comments.append([g['block'][2:-2], self.line_num])
                self.line_num += g['comments'].count('\n')
            elif self.lines:
                if g['start'] is None:
                    self.line_comments.append([g['line'][2:], self.line_num])
                    self.line_num += g['comments'].count('\n')
                else:
                    # Cosecutive lines with only comments with same leading whitespace
                    # will be captured as a single block.
                    if (
                        self.group_comments and
                        self.line_num == self.prev_line + 1 and
                        g['leading_space'] == self.leading
                    ):
                        self.line_comments[-1][0] += '\n' + g['line'][2:]
                    else:
                        self.line_comments.append([g['line'][2:], self.line_num])
                    self.leading = g['leading_space']
                    self.line_num += g['comments'].count('\n')
                    self.prev_line = self.line_num
            else:
                self.line_num += g['comments'].count('\n')
            text = ''.join([x[0] for x in RE_LINE_PRESERVE.findall(g["comments"])])
        return text

    def find_comments(self, text):
        """Find comments."""

        return ''.join(map(lambda m: self._evaluate(m), RE_COMMENT.finditer(text)))

    def _filter(self, text, context, encoding):
        """Filter JavaScript comments."""

        content = []
        self.line_num = 1
        self.prev_line = -1
        self.leading = ''
        self.jsdoc_comments = []
        self.block_comments = []
        self.line_comments = []

        self.find_comments(text)
        for comment, line in self.jsdoc_comments:
            content.append(
                filters.SourceText(
                    textwrap.dedent(comment),
                    "%s (%d)" % (context, line),
                    encoding,
                    'jsdocs'
                )
            )
        for comment, line in self.block_comments:
            content.append(
                filters.SourceText(
                    textwrap.dedent(comment),
                    "%s (%d)" % (context, line),
                    encoding,
                    'block-comment'
                )
            )
        for comment, line in self.line_comments:
            content.append(
                filters.SourceText(
                    textwrap.dedent(comment),
                    "%s (%d)" % (context, line),
                    encoding,
                    'line-comment'
                )
            )

        return content

    def filter(self, source_file, encoding):  # noqa A001
        """Parse HTML file."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            text = f.read()

        return self._filter(text, source_file, encoding)

    def sfilter(self, source):
        """Filter."""

        return self._filter(source.text, source.context, source.encoding)


def get_filter():
    """Get filter."""

    return JavaScriptFilter



"""
Python filter.

Parse Python docstrings.
"""
from __future__ import unicode_literals
from .. import filters
import re
import textwrap
import tokenize
import codecs
import io

tokenizer = tokenize.generate_tokens
PREV_DOC_TOKENS = (tokenize.INDENT, tokenize.DEDENT, tokenize.NEWLINE, tokenize.ENCODING)

RE_PY_ENCODE = re.compile(
    br'^[^\r\n]*?coding[:=]\s*([-\w.]+)|[^\r\n]*?\r?\n[^\r\n]*?coding[:=]\s*([-\w.]+)'
)
RE_NON_PRINTABLE_ASCII = re.compile(br"[^ -~]+")

DEFAULT_ENCODING = 'utf-8'


class PythonFilter(filters.Filter):
    """Spelling Python."""

    MODULE = 0
    FUNCTION = 1
    CLASS = 2

    def __init__(self, options, default_encoding=DEFAULT_ENCODING):
        """Initialization."""

        self.comments = options.get('comments', True) is True
        self.docstrings = options.get('docstrings', True) is True
        self.strings = options.get('strings', False) is True
        self.bytes = options.get('bytes', False) is True
        self.group_comments = options.get('group_comments', False) is True
        super(PythonFilter, self).__init__(options, default_encoding)

    def header_check(self, content):
        """Special Python encoding check."""

        encode = None

        m = RE_PY_ENCODE.match(content)
        if m:
            if m.group(1):
                encode = m.group(1).decode('ascii')
            elif m.group(2):
                encode = m.group(2).decode('ascii')
        if encode is None:
            encode = DEFAULT_ENCODING
        return encode

    def get_ascii(self, text):
        """Retrieve ASCII text from byte string."""

        return RE_NON_PRINTABLE_ASCII.sub(r' ', text).decode('ascii')

    def _filter(self, text, context, encoding):
        """Retrieve the Python docstrings."""

        docstrings = []
        comments = []
        strings = []
        prev_token_type = tokenize.NEWLINE
        indent = ''
        name = None
        stack = [(context, 0, self.MODULE)]

        src = io.StringIO(text)

        for token in tokenizer(src.readline):
            token_type = token[0]
            value = token[1]
            line = str(token[2][0])
            line_num = token[2][0]

            value = token[1]
            line = str(token[2][0])

            # Track function and class ancestry
            if token_type == tokenize.NAME:
                if value in ('def', 'class'):
                    name = value
                elif name:
                    parent = stack[-1][2]
                    prefix = ''
                    if parent != self.MODULE:
                        prefix = '.' if parent == self.CLASS else ', '
                    if name == 'class':
                        stack.append(('%s%s' % (prefix, value), len(indent), self.CLASS))
                    elif name == 'def':
                        stack.append(('%s%s()' % (prefix, value), len(indent), self.FUNCTION))
                    name = None

            if token_type == tokenize.COMMENT and self.comments:
                # Capture comments
                if (
                    self.group_comments and
                    prev_token_type == tokenize.NL and
                    comments and (comments[-1][2] + 1) == line_num
                ):
                    # Group multiple consecutive comments
                    comments[-1][0] += '\n' + value[1:]
                    comments[-1][2] = line_num
                else:
                    if len(stack) > 1:
                        loc = "%s(%s): %s" % (stack[0][0], line, ''.join([crumb[0] for crumb in stack[1:]]))
                    else:
                        loc = "%s(%s)" % (stack[0][0], line)
                    comments.append([value[1:], loc, line_num])
            if token_type == tokenize.STRING:
                # Capture docstrings
                # If previously we captured an INDENT or NEWLINE previously we probably have a docstring.
                # NL seems to be a different thing.
                if prev_token_type in PREV_DOC_TOKENS:
                    if self.docstrings:
                        value = value.strip()
                        string = textwrap.dedent(eval(value))

                        if not isinstance(string, str):
                            # Since docstrings should be readable and printable,
                            # if byte string assume 'ascii'.
                            string = string.decode('ascii')
                        loc = "%s(%s): %s" % (stack[0][0], line, ''.join([crumb[0] for crumb in stack[1:]]))
                        docstrings.append(filters.SourceText(string, loc, encoding, 'docstring'))
                elif self.strings:
                    value = value.strip()
                    string = textwrap.dedent(eval(value))
                    if isinstance(string, str) or self.bytes:
                        string_type = 'string'
                        if not isinstance(string, str):
                            string = self.get_ascii(string)
                            string_type = 'bytes'
                        loc = "%s(%s): %s" % (stack[0][0], line, ''.join([crumb[0] for crumb in stack[1:]]))
                        strings.append(filters.SourceText(string, loc, encoding, string_type))

            if token_type == tokenize.INDENT:
                indent = value
            elif token_type == tokenize.DEDENT:
                indent = indent[:-4]
                if len(stack) > 1 and len(indent) <= stack[-1][1]:
                    stack.pop()

            prev_token_type = token_type

        final_comments = []
        for comment in comments:
            final_comments.append(filters.SourceText(textwrap.dedent(comment[0]), comment[1], encoding, 'comment'))

        return docstrings + final_comments + strings

    def filter(self, source_file, encoding):  # noqa A001
        """Parse Python file returning content."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            return self._filter(f.read(), source_file, encoding)

    def sfilter(self, source):
        """Filter."""

        return self._filter(source.text, source.context, source.encoding)


def get_filter():
    """Return the filter."""

    return PythonFilter



"""Spell check with Aspell or Hunspell."""
from __future__ import unicode_literals
import os
import codecs
import importlib
from . import util
from . import __version__
from . import settings
from wcmatch import glob

version = __version__.version
version_info = __version__.version_info

__all__ = ("version", "version_info", "spellcheck")


class SpellChecker(object):
    """Spell check class."""

    DICTIONARY = 'dictionary.dic'

    GLOB_FLAG_MAP = {
        "FORECECASE": glob.F,
        "F": glob.F,
        "IGNORECASE": glob.I,
        "I": glob.I,
        "RAWCHARS": glob.R,
        "R": glob.R,
        "NEGATE": glob.N,
        "N": glob.N,
        "MINUSNEGATE": glob.M,
        "M": glob.M,
        "GLOBSTAR": glob.G,
        "G": glob.G,
        "DOTGLOB": glob.D,
        "D": glob.D,
        "EXTGLOB": glob.E,
        "E": glob.E,
        "BRACE": glob.B,
        "B": glob.B
    }

    def __init__(self, config, binary='', verbose=0):
        """Initialize."""

        # General options
        self.binary = binary if binary else 'aspell'
        self.verbose = verbose
        self.dict_bin = os.path.abspath(self.DICTIONARY)

    def log(self, text, level):
        """Log level."""
        if self.verbose >= level:
            print(text)

    def _normalize_utf(self, encoding):
        """Normalize UTF encoding."""

        if encoding == 'utf-8-sig':
            encoding = 'utf-8'
        if encoding.startswith('utf-16'):
            encoding = 'utf-16'
        elif encoding.startswith('utf-32'):
            encoding = 'utf-32'
        return encoding

    def setup_command(self, encoding, options, personal_dict):
        """Setup the command."""

        return []

    def _check_source(self, sources, options, personal_dict, filter_index=1):
        """Recursive check spelling filters."""
        for source in sources:
            if source._has_error():
                yield source
            elif not source._is_bytes():
                if filter_index < len(self.filters):
                    f = self.filters[filter_index]
                    if source.category not in f._get_disallowed():
                        yield from self._check_source(
                            f.sfilter(source), options, personal_dict, filter_index + 1
                        )
                    else:
                        yield source
                else:
                    yield source
            else:
                yield source

    def _check_spelling(self, sources, options, personal_dict):
        """Check spelling."""

        for source in self._check_source(sources, options, personal_dict):
            if source._has_error():
                yield util.Results([], source.context, source.category, source.error)
            else:
                if source._is_bytes():
                    text = source.text
                else:
                    text = source.text.encode(self._normalize_utf(source.encoding))
                self.log(text, 3)
                cmd = self.setup_command(self._normalize_utf(source.encoding), options, personal_dict)
                self.log(str(cmd), 2)

                wordlist = util.call_spellchecker(cmd, input_text=text)
                yield util.Results([w for w in sorted(set(wordlist.split('\n'))) if w], source.context, source.category)

    def compile_dictionary(self, lang, wordlists, output):
        """Compile user dictionary."""

    def _walk_src(self, targets, flags, plugin):
        """Walk source and parse files."""

        for target in targets:
            patterns = glob.globsplit(target, flags=flags)
            for f in glob.iglob(patterns, flags=flags):
                if not os.path.isdir(f):
                    yield plugin._parse(f)

    def setup_spellchecker(self, documents):
        """Setup spell checker."""

        return {}

    def setup_dictionary(self, documents):
        """Setup dictionary."""

        return None

    def _get_filters(self, documents, default_encoding):
        """Get filters."""

        self.filters = []
        kwargs = {}
        if default_encoding:
            kwargs["default_encoding"] = default_encoding
        filters = documents.get('filters', [])
        if not filters:
            filters.append('pyspelling.filters.text')
        for f in filters:
            # Retrieve module and module options
            if isinstance(f, dict):
                name, options = list(f.items())[0]
            else:
                name = f
                options = {}
            if options is None:
                options = {}

            self.filters.append(self._get_module(name, 'get_filter')(options, **kwargs))

    def _get_module(self, module, accessor):
        """Get module."""

        if isinstance(module, str):
            mod = importlib.import_module(module)
        attr = getattr(mod, accessor, None)
        if not attr:
            raise ValueError("Could not find accessor '%s'!" % accessor)
        return attr()

    def _to_flags(self, text):
        """Convert text representation of flags to actual flags."""

        flags = 0
        for x in text.split('|'):
            value = x.strip().upper()
            if value:
                flags |= self.GLOB_FLAG_MAP.get(value, 0)
        return flags

    def check(self, documents):
        """Walk source and initiate spell check."""

        # Perform spell check
        self.log('Spell Checking %s...' % documents.get('name', ''), 1)

        # Setup filters and variables for the spell check
        encoding = documents.get('default_encoding', '')
        options = self.setup_spellchecker(documents)
        output = self.setup_dictionary(documents)
        glob_flags = self._to_flags(documents.get('glob_flags', "N|B|G"))
        self._get_filters(documents, encoding)

        for sources in self._walk_src(documents.get('sources', []), glob_flags, self.filters[0]):
            yield from self._check_spelling(sources, options, output)


class Aspell(SpellChecker):
    """Aspell spell check class."""

    def __init__(self, config, binary='', verbose=0):
        """Initialize."""

        super(Aspell, self).__init__(config, binary, verbose)
        self.binary = binary if binary else 'aspell'

    def setup_spellchecker(self, documents):
        """Setup spell checker."""

        return documents.get('aspell', {})

    def setup_dictionary(self, documents):
        """Setup dictionary."""

        dictionary_options = documents.get('dictionary', {})
        output = os.path.abspath(dictionary_options.get('output', self.dict_bin))
        lang = dictionary_options.get('lang', 'en')
        wordlists = dictionary_options.get('wordlists', [])
        if lang and wordlists:
            self.compile_dictionary(lang, dictionary_options.get('wordlists', []), output)
        else:
            output = None
        return output

    def compile_dictionary(self, lang, wordlists, output):
        """Compile user dictionary."""

        output_location = os.path.dirname(output)
        if not os.path.exists(output_location):
            os.makedirs(output_location)
        if os.path.exists(output):
            os.remove(output)

        self.log("Compiling Dictionary...", 1)
        # Read word lists and create a unique set of words
        words = set()
        for wordlist in wordlists:
            with open(wordlist, 'rb') as src:
                for word in src.read().split(b'\n'):
                    words.add(word.replace(b'\r', b''))

        # Compile wordlist against language
        util.call(
            [
                self.binary,
                '--lang', lang,
                '--encoding=utf-8',
                'create',
                'master', output
            ],
            input_text=b'\n'.join(sorted(words)) + b'\n'
        )

    def setup_command(self, encoding, options, personal_dict):
        """Setup the command."""

        cmd = [
            self.binary,
            'list',
            '--encoding', codecs.lookup(encoding).name
        ]

        if personal_dict:
            cmd.extend(['--add-extra-dicts', personal_dict])

        allowed = {
            'conf-dir', 'data-dir', 'add-dict-alias', 'rem-dict-alias', 'dict-dir',
            'encoding', 'add-filter', 'rem-filter', 'add-filter-path', 'rem-filter-path',
            'mode', 'e', 'H', 't', 'n', 'add-extra-dicts', 'rem-extra-dicts', 'home-dir',
            'ingore', 'W', 'dont-ignore-case', 'ignore-case', 'lang', 'l', 'local-data-dir',
            'd', 'master', 'dont-normalize', 'normalize', 'dont-norm-required',
            'norm-required', 'norm-form', 'dont-norm-strict', 'norm-strict', 'per-conf',
            'p', 'personal', 'C', 'B', 'dont-run-together', 'run-together', 'run-together-limit',
            'run-together-min', 'use-other-dicts', 'dont-use-other-dicts', 'add-variety', 'rem-variety',
            'add-context-delimiters', 'rem-context-delimiters', 'dont-context-visible-first',
            'context-visible-first', 'add-email-quote', 'rem-email-quote', 'email-margin',
            'add-html-check', 'rem-html-check', 'add-html-skip', 'rem-html-skip', 'add-sgml-check',
            'rem-sgml-check', 'add-sgml-skip', 'rem-sgml-skip', 'dont-tex-check-comments',
            'tex-check-comments', 'add-tex-command', 'rem-tex-command', 'add-texinfo-ignore',
            'rem-texinfo-ignore', 'add-texinfo-ignore-env', 'rem-texinfo-ignore-env', 'filter'
        }

        for k, v in options.items():
            if k in allowed:
                key = ('-%s' if len(k) == 1 else '--%s') % k
                if isinstance(v, bool) and v is True:
                    cmd.append(key)
                elif isinstance(v, str):
                    cmd.extend([key, v])
                elif isinstance(v, int):
                    cmd.extend([key, str(v)])
                elif isinstance(v, list):
                    for value in v:
                        cmd.extend([key, str(value)])
        return cmd


class Hunspell(SpellChecker):
    """Hunspell spell check class."""

    def __init__(self, config, binary='', verbose=0):
        """Initialize."""

        super(Hunspell, self).__init__(config, binary, verbose)
        self.binary = binary if binary else 'hunspell'

    def setup_spellchecker(self, documents):
        """Setup spell checker."""

        return documents.get('hunspell', {})

    def setup_dictionary(self, documents):
        """Setup dictionary."""

        dictionary_options = documents.get('dictionary', {})
        output = os.path.abspath(dictionary_options.get('output', self.dict_bin))
        lang = dictionary_options.get('lang', 'en_US')
        wordlists = dictionary_options.get('wordlists', [])
        if lang and wordlists:
            self.compile_dictionary(lang, dictionary_options.get('wordlists', []), output)
        else:
            output = None
        return output

    def compile_dictionary(self, lang, wordlists, output):
        """Compile user dictionary."""

        output_location = os.path.dirname(output)
        if not os.path.exists(output_location):
            os.makedirs(output_location)
        if os.path.exists(output):
            os.remove(output)

        self.log("Compiling Dictionary...", 1)
        # Read word lists and create a unique set of words
        words = set()
        for wordlist in wordlists:
            with open(wordlist, 'rb') as src:
                for word in src.read().split(b'\n'):
                    words.add(word.replace(b'\r', b''))

        # Sort and create wordlist
        with open(output, 'wb') as dest:
            dest.write(b'\n'.join(sorted(words)) + b'\n')

    def setup_command(self, encoding, options, personal_dict):
        """Setup command."""

        cmd = [
            self.binary,
            '-l',
            '-i', codecs.lookup(encoding).name
        ]

        if personal_dict:
            cmd.extend(['-p', personal_dict])

        allowed = {
            'check-apostrophe', 'check-url',
            'd', 'H', 'n', 'o', 'r', 't', 'X'
        }

        for k, v in options.items():
            if k in allowed:
                key = ('-%s' if len(k) == 1 else '--%s') % k
                if isinstance(v, bool) and v is True:
                    cmd.append(key)
                elif isinstance(v, str):
                    cmd.extend([key, v])
                elif isinstance(v, int):
                    cmd.extend([key, str(v)])
                elif isinstance(v, list):
                    for value in v:
                        cmd.extend([key, str(value)])
        return cmd


def spellcheck(config_file, name='', binary='', verbose=0, checker=''):
    """Spell check."""

    hunspell = None
    aspell = None
    spellchecker = None
    config = settings.read_config(config_file)

    for documents in config.get('documents', []):

        if name and name != documents.get('name', ''):
            continue

        if checker == "hunspell":
            if hunspell is None:
                hunspell = Hunspell(config, binary, verbose)
            spellchecker = hunspell

        elif checker == "aspell":
            if aspell is None:
                aspell = Aspell(config, binary, verbose)
            spellchecker = aspell
        else:
            raise ValueError('%s is not a valid spellchecker!' % checker)

        spellchecker.log('==> Using %s to spellcheck %s' % (checker, documents.get('name', '')), 1)
        yield from spellchecker.check(documents)
        spellchecker.log("", 1)



"""Settings."""
from __future__ import unicode_literals
import yaml
import codecs

__all__ = ("yaml_load", "read_config")


def yaml_load(source, loader=yaml.Loader):
    """
    Wrap PyYaml's loader so we can extend it to suit our needs.

    Load all strings as Unicode: http://stackoverflow.com/a/2967461/3609487.
    """

    def construct_yaml_str(self, node):
        """Override the default string handling function to always return Unicode objects."""
        return self.construct_scalar(node)

    class Loader(loader):
        """Define a custom loader to leave the global loader unaltered."""

    # Attach our Unicode constructor to our custom loader ensuring all strings
    # will be Unicode on translation.
    Loader.add_constructor('tag:yaml.org,2002:str', construct_yaml_str)

    return yaml.load(source, Loader)


def read_config(file_name):
    """Read configuration."""

    config = {}
    with codecs.open(file_name, 'r', encoding='utf-8') as f:
        config = yaml_load(f.read())
    return config



"""`PySpelling` package."""

#   (major, minor, micro, release type, pre-release build, post-release build)
version_info = (0, 2, 0, 'alpha', 2, 0)


def _version():
    """
    Get the version (PEP 440).

    Version structure
      (major, minor, micro, release type, pre-release build, post-release build)
    Release names are named is such a way they are sortable and comparable with ease.
      (alpha | beta | candidate | final)

    - "final" should never have a pre-release build number
    - pre-releases should have a pre-release build number greater than 0
    - post-release is only applied if post-release build is greater than 0
    """

    releases = {"alpha": 'a', "beta": 'b', "candidate": 'rc', "final": ''}
    # Version info should be proper length
    assert len(version_info) == 6
    # Should be a valid release
    assert version_info[3] in releases
    # Pre-release releases should have a pre-release value
    assert version_info[3] == 'final' or version_info[4] > 0
    # Final should not have a pre-release value
    assert version_info[3] != 'final' or version_info[4] == 0

    main = '.'.join(str(x)for x in (version_info[0:2] if version_info[2] == 0 else version_info[0:3]))
    prerel = releases[version_info[3]]
    prerel += str(version_info[4]) if prerel else ''
    postrel = '.post%d' % version_info[5] if version_info[5] > 0 else ''

    return ''.join((main, prerel, postrel))


version = _version()
